from thinc.api import Model
from typing import List
from ..data_classes import FullTransformerBatch, TransformerData


def split_trf_batch() -> Model[FullTransformerBatch, List[TransformerData]]:
    return Model("split-trf-batch", forward)


def forward(model, trf_full, is_train):
    def backprop(d_trf_datas):
        return trf_full.unsplit_by_doc([x.tensors for x in d_trf_datas])

    return trf_full.doc_data, backprop



import spacy
from spacy.training.example import Example
from spacy.util import make_tempdir
from spacy import util
from thinc.api import Config


TRAIN_DATA = [
    ("I'm so happy.", {"cats": {"POSITIVE": 1.0, "NEGATIVE": 0.0}}),
    ("I'm so angry", {"cats": {"POSITIVE": 0.0, "NEGATIVE": 1.0}}),
]


cfg_string = """
    [nlp]
    lang = "en"
    pipeline = ["transformer","textcat"]

    [components]

    [components.textcat]
    factory = "textcat"

    [components.textcat.model]
    @architectures = "spacy.TextCatEnsemble.v2"

    [components.textcat.model.tok2vec]
    @architectures = "spacy-transformers.TransformerListener.v1"
    grad_factor = 1.0

    [components.textcat.model.tok2vec.pooling]
    @layers = "reduce_mean.v1"

    [components.transformer]
    factory = "transformer"
    """


def test_transformer_pipeline_textcat():
    """Test that a pipeline with just a transformer+textcat runs and trains properly.
    This used to throw an error because of shape inference issues -
    cf https://github.com/explosion/spaCy/issues/6401"""
    orig_config = Config().from_str(cfg_string)
    nlp = util.load_model_from_config(orig_config, auto_fill=True, validate=True)
    assert nlp.pipe_names == ["transformer", "textcat"]
    train_examples = []

    for text, annotations in TRAIN_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(text), annotations))
    optimizer = nlp.initialize(get_examples=lambda: train_examples)

    for i in range(2):
        losses = {}
        nlp.update(train_examples, sgd=optimizer, losses=losses)

    doc = nlp("We're interested at underwater basket weaving.")
    cats1 = doc.cats

    # ensure IO goes OK
    with make_tempdir() as d:
        file_path = d / "trained_nlp"
        nlp.to_disk(file_path)
        nlp2 = spacy.load(file_path)
        doc2 = nlp2("We're interested at underwater basket weaving.")
        cats2 = doc2.cats
        assert cats1 == cats2



from dataclasses import dataclass, field
from typing import Optional, List, Dict, Any, Union, Tuple
import torch
import numpy
from transformers.tokenization_utils import BatchEncoding
from transformers.file_utils import ModelOutput
from thinc.types import Ragged, Floats2d, Floats3d, FloatsXd, Ints2d
from thinc.api import NumpyOps, get_array_module, xp2torch, torch2xp
from spacy.tokens import Span
import srsly

from .util import transpose_list
from .align import get_token_positions


@dataclass
class WordpieceBatch:
    """Holds data from the transformers BatchEncoding class.

    We would have preferred to use the BatchEncoding class directly, but
    there's a few problems with that.

    1. Some BatchEncoding functionality requires the tokenizers.Encoding object,
        and it's impossible for us to create or manipulate that object. This means
        we can't really create BatchEncoding objects, which limits what we can do.
    2. We want some semantic differences, for instance the "lengths" data in the
        BatchEncoding is useless when the inputs are padded. We want it to tell
        us the *unpadded* lengths.
    3. We want typed attributes, so that we can type-check properly.
    4. We prefer to have numpy/cupy arrays rather than torch arrays.
    5. The API around the BatchEncoding object has been changing a lot, so we
        want to minimize the places where we touch it.
    """

    strings: List[List[str]]
    input_ids: Ints2d
    attention_mask: Floats2d
    lengths: List[int]
    token_type_ids: Optional[Ints2d]

    def __len__(self) -> int:
        return len(self.strings)

    def __getitem__(self, index) -> "WordpieceBatch":
        if isinstance(index, int):
            slice_ = slice(index, index + 1)
        else:
            slice_ = index
        return WordpieceBatch(
            strings=self.strings[slice_],
            input_ids=self.input_ids[slice_],
            attention_mask=self.attention_mask[slice_],
            lengths=self.lengths[slice_],
            token_type_ids=(
                self.token_type_ids[slice_] if self.token_type_ids is not None else None
            ),
        )

    def to_hf_dict(self) -> Dict:
        """Return a dict similar to the format produced by the Huggingface
        tokenizer, converting arrays to pytorch tensors as well.
        """
        output = {
            "input_ids": xp2torch(self.input_ids),
            "attention_mask": xp2torch(self.attention_mask),
            "input_texts": self.strings,
        }
        if self.token_type_ids is not None:
            output["token_type_ids"] = xp2torch(self.token_type_ids)
        return output

    @classmethod
    def empty(cls, *, xp=numpy) -> "WordpieceBatch":
        return cls(
            strings=[],
            input_ids=xp.zeros((0, 0), dtype="i"),
            attention_mask=xp.ones((0, 0), dtype="bool"),
            lengths=[],
            token_type_ids=None,
        )

    @classmethod
    def zeros(cls, lengths: List[int], xp=numpy) -> "WordpieceBatch":
        return cls(
            strings=[[""] * length for length in lengths],
            input_ids=xp.array([[0] * length for length in lengths], dtype="i"),
            attention_mask=xp.ones((len(lengths), max(lengths)), dtype="bool"),
            lengths=lengths,
            token_type_ids=None,
        )

    @classmethod
    def from_batch_encoding(cls, token_data: BatchEncoding) -> "WordpieceBatch":
        assert isinstance(token_data, BatchEncoding) or isinstance(token_data, dict)
        pad_token = token_data.get("pad_token", "[PAD]")
        lengths = [
            len([tok for tok in tokens if tok != pad_token])
            for tokens in token_data["input_texts"]
        ]

        numpy_ops = NumpyOps()

        return cls(
            strings=token_data["input_texts"],
            input_ids=numpy_ops.asarray2i(token_data["input_ids"]),
            attention_mask=numpy_ops.asarray2f(token_data["attention_mask"]),
            lengths=lengths,
            token_type_ids=(
                numpy_ops.asarray2i(token_data["token_type_ids"])
                if "token_type_ids" in token_data
                else None
            ),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            "strings": self.strings,
            "input_ids": self.input_ids,
            "attention_mask": self.attention_mask,
            "lengths": self.lengths,
            "token_type_ids": self.token_type_ids,
        }

    def from_dict(self, msg: Dict[str, Any]) -> "WordpieceBatch":
        self.strings = msg["strings"]
        self.input_ids = msg["input_ids"]
        self.attention_mask = msg["attention_mask"]
        self.lengths = msg["lengths"]
        self.token_type_ids = msg["token_type_ids"]
        return self


@dataclass
class TransformerData:
    """Transformer tokens and outputs for one Doc object.

    The transformer models return tensors that refer to a whole padded batch
    of documents. These tensors are wrapped into the FullTransformerBatch object.
    The FullTransformerBatch then splits out the per-document data, which is
    handled by this class. Instances of this class are typically assigned to
    the doc._.trf_data extension attribute.

    Attributes
    ----------
    wordpieces (WordpieceBatch): A slice of the wordpiece token data produced
        by the Huggingface tokenizer.
    model_output (ModelOutput): The model output from the transformer model,
        determined by the model and transformer config.
    align (Ragged): Alignment from the Doc's tokenization to the wordpieces.
        This is a ragged array, where align.lengths[i] indicates the number of
        wordpiece tokens that token i aligns against. The actual indices are
        provided at align[i].dataXd.
    """

    wordpieces: WordpieceBatch
    model_output: ModelOutput
    align: Ragged

    @classmethod
    def empty(cls) -> "TransformerData":
        align = Ragged(numpy.zeros((0,), dtype="i"), numpy.zeros((0,), dtype="i"))
        return cls(
            wordpieces=WordpieceBatch.empty(), model_output=ModelOutput(), align=align
        )

    @classmethod
    def zeros(cls, length: int, width: int, *, xp=numpy) -> "TransformerData":
        """Create a valid TransformerData container for a given shape, filled
        with zeros."""
        return cls(
            wordpieces=WordpieceBatch.zeros([length], xp=xp),
            model_output=ModelOutput(
                last_hidden_state=xp.zeros((1, length, width), dtype="f")
            ),
            align=Ragged(numpy.arange(length), numpy.ones((length,), dtype="i")),
        )

    @property
    def tensors(self) -> Tuple[Union[FloatsXd, List[FloatsXd]]]:
        return self.model_output.to_tuple()

    @property
    def tokens(self) -> Dict[str, Any]:
        """Deprecated. A dict with the wordpiece token data."""
        return self.wordpieces.to_hf_dict()

    @property
    def width(self) -> int:
        if "last_hidden_state" in self.model_output:
            return self.model_output.last_hidden_state.shape[-1]
        else:
            raise ValueError("Cannot find last hidden state")

    def to_dict(self) -> Dict[str, Any]:
        return {
            "wordpieces": self.wordpieces.to_dict(),
            "model_output": self.model_output,
            "align": [self.align.dataXd, self.align.lengths],
        }

    def from_dict(self, msg: Dict[str, Any]) -> "TransformerData":
        self.wordpieces = WordpieceBatch.empty().from_dict(msg["wordpieces"])
        self.model_output = ModelOutput(msg["model_output"])
        self.align = Ragged(*msg["align"])
        return self

    def to_bytes(self) -> bytes:
        return srsly.msgpack_dumps(self.to_dict())

    def from_bytes(self, byte_string: bytes) -> "TransformerData":
        msg = srsly.msgpack_loads(byte_string)
        self.from_dict(msg)
        return self


@srsly.msgpack_encoders("transformerdata")
def serialize_transformer_data(obj, chain=None):
    if isinstance(obj, TransformerData):
        return {"__transformerdata__": obj.to_dict()}
    return obj if chain is None else chain(obj)


@srsly.msgpack_decoders("transformerdata")
def deserialize_transformer_data(obj, chain=None):
    if "__transformerdata__" in obj:
        return TransformerData.empty().from_dict(obj["__transformerdata__"])
    return obj if chain is None else chain(obj)


@dataclass
class FullTransformerBatch:
    """Holds a batch of input and output objects for a transformer model. The
    data can then be split to a list of `TransformerData` objects to associate
    the outputs to each `Doc` in the batch.

    Attributes
    ----------
    spans (List[List[Span]]): The batch of input spans. The outer list refers
        to the Doc objects in the batch, and the inner list are the spans for
        that `Doc`. Note that spans are allowed to overlap or exclude tokens,
        but each Span can only refer to one Doc (by definition). This means that
        within a Doc, the regions of the output tensors that correspond to each
        Span may overlap or have gaps, but for each Doc, there is a non-overlapping
        contiguous slice of the outputs.
    wordpieces (WordpieceBatch): Token data from the Huggingface tokenizer.
    model_output (ModelOutput): The output of the transformer model.
    align (Ragged): Alignment from the spaCy tokenization to the wordpieces.
        This is a ragged array, where align.lengths[i] indicates the number of
        wordpiece tokens that token i aligns against. The actual indices are
        provided at align[i].dataXd.
    """

    spans: List[List[Span]]
    wordpieces: WordpieceBatch
    model_output: ModelOutput
    align: Ragged
    cached_doc_data: Optional[List[TransformerData]] = None

    @classmethod
    def empty(cls, nr_docs) -> "FullTransformerBatch":
        spans = [[] for _ in range(nr_docs)]
        doc_data = [TransformerData.empty() for _ in range(nr_docs)]
        align = Ragged(numpy.zeros((0,), dtype="i"), numpy.zeros((0,), dtype="i"))
        return cls(
            spans=spans,
            wordpieces=WordpieceBatch.empty(),
            model_output=ModelOutput(),
            align=align,
            cached_doc_data=doc_data,
        )

    @property
    def tensors(self) -> Tuple[Union[torch.Tensor, Tuple[torch.Tensor]]]:
        return self.model_output.to_tuple()

    @property
    def tokens(self) -> Dict[str, Any]:
        """Deprecated. Dict formatted version of the self.wordpieces data,
        with values converted to PyTorch tensors.
        """
        return self.wordpieces.to_hf_dict()

    @property
    def doc_data(self) -> List[TransformerData]:
        """The outputs, split per spaCy Doc object."""
        if self.cached_doc_data is None:
            self.cached_doc_data = self.split_by_doc()
        return self.cached_doc_data

    def unsplit_by_doc(self, arrays: List[List[Floats3d]]) -> "FullTransformerBatch":
        """Return a new FullTransformerBatch from a split batch of activations,
        using the current object's spans, wordpieces and alignment.

        This is used during the backward pass, in order to construct the gradients
        to pass back into the transformer model.
        """
        xp = get_array_module(arrays[0][0])
        # construct a dummy ModelOutput with the tensor values
        model_output = ModelOutput()
        for i, x in enumerate(transpose_list(arrays)):
            model_output[f"output_{i}"] = xp2torch(xp.vstack(x))
        return FullTransformerBatch(
            spans=self.spans,
            wordpieces=self.wordpieces,
            model_output=model_output,
            align=self.align,
        )

    def split_by_doc(self) -> List[TransformerData]:
        """Split a TransformerData that represents a batch into a list with
        one TransformerData per Doc.
        """
        flat_spans = []
        for doc_spans in self.spans:
            flat_spans.extend(doc_spans)
        token_positions = get_token_positions(flat_spans)
        outputs = []
        start = 0
        prev_tokens = 0
        for doc_spans in self.spans:
            if len(doc_spans) == 0 or len(doc_spans[0]) == 0:
                outputs.append(TransformerData.empty())
                continue
            start_i = token_positions[doc_spans[0][0]]
            end_i = token_positions[doc_spans[-1][-1]] + 1
            end = start + len(doc_spans)
            doc_tokens = self.wordpieces[start:end]
            doc_align = self.align[start_i:end_i]
            doc_align.data = doc_align.data - prev_tokens
            model_output = ModelOutput()
            last_hidden_state = self.model_output.last_hidden_state
            for key, output in self.model_output.items():
                if isinstance(output, torch.Tensor):
                    model_output[key] = torch2xp(output[start:end])
                elif (
                    isinstance(output, tuple)
                    and all(isinstance(t, torch.Tensor) for t in output)
                    and all(t.shape[0] == last_hidden_state.shape[0] for t in output)
                ):
                    model_output[key] = [torch2xp(t[start:end]) for t in output]
            outputs.append(
                TransformerData(
                    wordpieces=doc_tokens,
                    model_output=model_output,
                    align=doc_align,
                )
            )
            prev_tokens += doc_tokens.input_ids.size
            start += len(doc_spans)
        return outputs


@dataclass
class HFObjects:

    tokenizer: Any
    transformer: Any
    vocab_file_contents: Any
    _init_tokenizer_config: Dict[str, Any] = field(default_factory=dict)
    _init_transformer_config: Dict[str, Any] = field(default_factory=dict)



from .listener import TransformerListener
from .transformer_model import TransformerModel
from .split_trf import split_trf_batch
from .trfs2arrays import trfs2arrays


__all__ = ["TransformerListener", "TransformerModel", "split_trf_batch", "trfs2arrays"]



import pytest
import numpy
from thinc.types import Ragged
from thinc.api import NumpyOps
from ..data_classes import WordpieceBatch
from ..truncate import _truncate_tokens, _truncate_alignment


@pytest.fixture
def sequences():
    # Each sequence is a list of tokens, and each token is a number of wordpieces
    return [
        [1, 3, 1],  # So 5 wordpieces this sequence
        [3, 7, 1, 1],  # 12
        [1],  # 1
        [20, 1],  # 21
    ]


@pytest.fixture
def shape(sequences):
    # Get the shape of the input_ids, which includes the padding.
    maximum = max(sum(lengths) for lengths in sequences)
    return (len(sequences), maximum)


@pytest.fixture
def seq_lengths(sequences):
    return numpy.array([sum(seq) for seq in sequences], dtype="i")


@pytest.fixture
def wordpieces(sequences):
    strings = []
    for token_lengths in sequences:
        strings.append([])
        for length in token_lengths:
            strings[-1].extend(str(i) for i in range(length))
    shape = (len(strings), max(len(seq) for seq in strings))
    wordpieces = WordpieceBatch(
        strings=strings,
        input_ids=numpy.zeros(shape, dtype="i"),
        token_type_ids=numpy.zeros(shape, dtype="i"),
        attention_mask=numpy.zeros((shape[0], shape[1]), dtype="bool"),
        lengths=[len(seq) for seq in strings],
    )
    return wordpieces


@pytest.fixture
def align(sequences):
    lengths = []
    indices = []
    offset = 0
    for seq in sequences:
        for token_length in seq:
            lengths.append(token_length)
            indices.extend(i + offset for i in range(token_length))
            offset += token_length
    return Ragged(numpy.array(indices, dtype="i"), numpy.array(lengths, dtype="i"))


@pytest.fixture
def max_length():
    return 6


@pytest.fixture
def mask_from_end(shape, max_length):
    n_seq, length = shape
    bools = [
        numpy.array([i < max_length for i in range(length)], dtype="bool")
        for _ in range(n_seq)
    ]
    return numpy.concatenate(bools)


def test_truncate_wordpieces(wordpieces, max_length, mask_from_end):
    truncated = _truncate_tokens(wordpieces, mask_from_end)
    for i, seq in enumerate(truncated.strings):
        assert len(seq) <= max_length
        assert seq == wordpieces.strings[i][:max_length]
        assert truncated.input_ids[i].shape[0] <= max_length
        assert truncated.token_type_ids[i].shape[0] <= max_length
        assert truncated.attention_mask[i].shape[0] <= max_length


def test_truncate_alignment_from_end(sequences, max_length, align, mask_from_end):
    # print("Max length", max_length)
    # print("Sequences", sequences)
    # print("Mask", mask_from_end)
    ops = NumpyOps()
    truncated = _truncate_alignment(align, mask_from_end)
    # print(truncated.dataXd.shape, truncated.lengths.sum())
    # print("Before", list(map(list, ops.unflatten(align.dataXd, align.lengths))))
    # print("After", list(map(list, ops.unflatten(truncated.dataXd, truncated.lengths))))
    # Check that the number of tokens hasn't changed. We still need to have
    # alignment for every token.
    assert truncated.lengths.shape[0] == align.lengths.shape[0]
    start = 0
    for i, seq in enumerate(sequences):
        end = start + len(seq)
        # Get the alignment for this sequence of tokens. Each length in the
        # alignment indicates the number of wordpiece tokens, so we need to
        # check that the sum of the lengths doesn't exceed the maximum.
        wp_indices = truncated[start:end]
        assert wp_indices.lengths.sum() <= max_length
        # We're truncating from the end, so we shouldn't see different values
        # except at the end of the sequence.
        seen_zero = False
        before = align[start:end]
        for length_now, length_before in zip(wp_indices.lengths, before.lengths):
            if seen_zero:
                assert length_now == 0, wp_indices.lengths
            elif length_now == 0:
                seen_zero = True
            else:
                length_now == length_before



from typing import Callable, Optional, Any
from thinc.layers.pytorchwrapper import forward as pt_forward
from thinc.layers.pytorchwrapper import convert_pytorch_default_inputs, convert_pytorch_default_outputs

from thinc.api import registry, Model

from .hf_shim import HFShim


@registry.layers("HFWrapper.v1")
def HFWrapper(
    hf_model: "HFObjects",
    convert_inputs: Optional[Callable] = None,
    convert_outputs: Optional[Callable] = None,
    mixed_precision: bool = False,
    grad_scaler_config: dict = {},
) -> Model[Any, Any]:
    """Wrap a PyTorch HF model, so that it has the same API as Thinc models.
    To optimize the model, you'll need to create a PyTorch optimizer and call
    optimizer.step() after each batch. See examples/wrap_pytorch.py

    Your PyTorch model's forward method can take arbitrary args and kwargs,
    but must return either a single tensor as output or a tuple. You may find the
    PyTorch register_forward_hook helpful if you need to adapt the output.

    The convert functions are used to map inputs and outputs to and from your
    PyTorch model. Each function should return the converted output, and a callback
    to use during the backward pass. So:

        Xtorch, get_dX = convert_inputs(X)
        Ytorch, torch_backprop = model.shims[0](Xtorch, is_train)
        Y, get_dYtorch = convert_outputs(Ytorch)

    To allow maximum flexibility, the PyTorchShim expects ArgsKwargs objects
    on the way into the forward and backward passed. The ArgsKwargs objects
    will be passed straight into the model in the forward pass, and straight
    into `torch.autograd.backward` during the backward pass.
    """
    if convert_inputs is None:
        convert_inputs = convert_pytorch_default_inputs
    if convert_outputs is None:
        convert_outputs = convert_pytorch_default_outputs

    return Model(
        "hf-pytorch",
        pt_forward,
        attrs={"convert_inputs": convert_inputs, "convert_outputs": convert_outputs},
        shims=[
            HFShim(
                hf_model,
                mixed_precision=mixed_precision,
                grad_scaler_config=grad_scaler_config,
            )
        ],
        dims={"nI": None, "nO": None},
    )



import pytest
from spacy.training.example import Example
from spacy.util import make_tempdir
from spacy import util
from thinc.api import Model, Config
from .util import _assert_equal_tensors

# fmt: off
TRAIN_DATA = [
    ("I like green eggs", {"tags": ["N", "V", "J", "N"]}),
    ("Eat blue ham", {"tags": ["V", "J", "N"]}),
]


cfg_string = """
    [nlp]
    lang = "en"
    pipeline = ["tagger"]

    [components]

    [components.tagger]
    factory = "tagger"

    [components.tagger.model]
    @architectures = "spacy.Tagger.v1"
    nO = null

    [components.tagger.model.tok2vec]
    @architectures = "spacy-transformers.Tok2VecTransformer.v1"
    name = "roberta-base"
    tokenizer_config = {"use_fast": false}
    grad_factor = 1.0

    [components.tagger.model.tok2vec.get_spans]
    @span_getters = "spacy-transformers.strided_spans.v1"
    window = 256
    stride = 256

    [components.tagger.model.tok2vec.pooling]
    @layers = "reduce_mean.v1"
    """
# fmt: on


def test_transformer_pipeline_tagger_internal():
    """Test that a tagger with internal transformer runs and trains properly"""
    orig_config = Config().from_str(cfg_string)
    nlp = util.load_model_from_config(orig_config, auto_fill=True, validate=True)
    assert nlp.pipe_names == ["tagger"]
    tagger = nlp.get_pipe("tagger")
    tagger_trf = tagger.model.get_ref("tok2vec").layers[0]
    assert isinstance(tagger_trf, Model)
    train_examples = []
    for t in TRAIN_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))
        for tag in t[1]["tags"]:
            tagger.add_label(tag)

    optimizer = nlp.initialize(lambda: train_examples)
    for i in range(2):
        losses = {}
        nlp.update(train_examples, sgd=optimizer, losses=losses)

    doc = nlp("We're interested at underwater basket weaving.")
    doc_tensor = tagger_trf.predict([doc])

    # ensure IO goes OK
    with make_tempdir() as d:
        file_path = d / "trained_nlp"
        nlp.to_disk(file_path)

        # results are not the same if we don't call from_disk
        nlp2 = util.load_model_from_config(orig_config, auto_fill=True, validate=True)
        nlp2.initialize(lambda: train_examples)
        doc2 = nlp2("We're interested at underwater basket weaving.")
        tagger2 = nlp2.get_pipe("tagger")
        tagger_trf2 = tagger2.model.get_ref("tok2vec").layers[0]
        doc_tensor2 = tagger_trf2.predict([doc2])
        with pytest.raises(AssertionError):
            _assert_equal_tensors(
                doc_tensor2.doc_data[0].tensors, doc_tensor.doc_data[0].tensors
            )

        # results ARE the same if we call from_disk
        nlp3 = util.load_model_from_config(orig_config, auto_fill=True, validate=True)
        nlp3.from_disk(file_path)
        doc3 = nlp3("We're interested at underwater basket weaving.")
        tagger3 = nlp3.get_pipe("tagger")
        tagger_trf3 = tagger3.model.get_ref("tok2vec").layers[0]
        doc_tensor3 = tagger_trf3.predict([doc3])
        _assert_equal_tensors(
            doc_tensor3.doc_data[0].tensors, doc_tensor.doc_data[0].tensors
        )



import pytest
import torch
from spacy.language import Language
from spacy.training.example import Example
from spacy.util import make_tempdir
from spacy.vocab import Vocab
from spacy.tokens import Doc
from spacy import util
from thinc.api import Model, Config, get_current_ops, NumpyOps
from spacy.tests.util import assert_docs_equal

from .util import DummyTransformer, _assert_equal_tensors
from .. import TransformerModel
from ..pipeline_component import Transformer
from ..layers import TransformerListener
from ..data_classes import TransformerData, FullTransformerBatch


torch.set_num_threads(1)


@pytest.fixture
def vocab():
    return Vocab()


@pytest.fixture
def docs(vocab):
    return [
        Doc(vocab, words=["hello", "world"]),
        Doc(vocab, words=["this", "is", "another"]),
    ]


@pytest.fixture
def component(vocab):
    return Transformer(Vocab(), DummyTransformer())


@pytest.fixture(scope="module")
def simple_nlp():
    nlp = Language()
    nlp.add_pipe("transformer")
    train_examples = []
    for t in TRAIN_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))

    optimizer = nlp.initialize()
    for i in range(2):
        losses = {}
        nlp.update(train_examples, sgd=optimizer, losses=losses)

    return nlp


def test_init(component):
    assert isinstance(component.vocab, Vocab)
    assert isinstance(component.model, Model)
    assert hasattr(component.set_extra_annotations, "__call__")
    assert component.listeners == []
    assert component.cfg == {"max_batch_items": 4096}


def test_predict(component, docs):
    trf_data = component.predict(docs)
    n_tokens = trf_data.wordpieces.input_ids.shape[1]
    width = component.model.layers[0].attrs["width"]
    assert isinstance(trf_data, FullTransformerBatch)
    assert (
        len(trf_data.model_output.last_hidden_state)
        == component.model.layers[0].attrs["depth"]
    )
    assert trf_data.model_output.last_hidden_state[0].shape == (
        len(docs),
        n_tokens,
        width,
    )


def test_set_annotations(component, docs):
    trf_data = component.predict(docs)
    component.set_annotations(docs, trf_data)
    for doc in docs:
        assert isinstance(doc._.trf_data, TransformerData)


def test_set_extra_annotations(component, docs):
    Doc.set_extension("custom_attr", default="")

    def custom_annotation_setter(docs, trf_data):
        doc_data = list(trf_data.doc_data)
        for doc, data in zip(docs, doc_data):
            doc._.custom_attr = data

    component.set_extra_annotations = custom_annotation_setter
    trf_data = component.predict(docs)
    component.set_annotations(docs, trf_data)
    for doc in docs:
        assert isinstance(doc._.custom_attr, TransformerData)


def test_listeners(component, docs):
    docs = list(component.pipe(docs))
    for listener in component.listeners:
        assert listener.verify_inputs(docs)


TRAIN_DATA = [
    (
        "I like green eggs",
        {"tags": ["N", "V", "J", "N"], "sent_starts": [True, False, True, False]},
    ),
    ("Eat blue ham", {"tags": ["V", "J", "N"], "sent_starts": [True, False, False]}),
]


def test_transformer_pipeline_simple(simple_nlp):
    """Test that a simple pipeline with just a transformer at least runs"""
    doc = simple_nlp("We're interested at underwater basket weaving.")
    assert doc


def test_transformer_pipeline_long_token(simple_nlp):
    """Test that a simple pipeline does not raise an error on texts that exceeds
    the model max length. We should truncate instead.
    """
    doc = simple_nlp("https://example.com/" + "a/" * 1000)
    assert len(doc) == 1


cfg_string = """
    [nlp]
    lang = "en"
    pipeline = ["transformer","tagger","senter"]

    [components]

    [components.senter]
    factory = "senter"

    [components.senter.model]
    @architectures = "spacy.Tagger.v1"
    nO = null

    [components.senter.model.tok2vec]
    @architectures = "spacy-transformers.TransformerListener.v1"
    grad_factor = 1.0
    upstream = ${components.transformer.name}

    [components.senter.model.tok2vec.pooling]
    @layers = "reduce_mean.v1"

    [components.tagger]
    factory = "tagger"

    [components.tagger.model]
    @architectures = "spacy.Tagger.v1"
    nO = null

    [components.tagger.model.tok2vec]
    @architectures = "spacy-transformers.TransformerListener.v1"
    grad_factor = 1.0
    upstream = ${components.transformer.name}

    [components.tagger.model.tok2vec.pooling]
    @layers = "reduce_mean.v1"

    [components.transformer]
    factory = "transformer"
    name = "custom_upstream"

    [components.transformer.model]
    @architectures = "spacy-transformers.TransformerModel.v3"
    name = "albert-base-v2"

    [components.transformer.model.transformer_config]
    output_attentions = true
    """


def test_transformer_pipeline_tagger_senter_listener():
    """Test that a pipeline with just a transformer+tagger+senter runs and
    trains properly"""
    orig_config = Config().from_str(cfg_string)
    nlp = util.load_model_from_config(orig_config, auto_fill=True, validate=True)
    assert nlp.pipe_names == ["transformer", "tagger", "senter"]
    tagger = nlp.get_pipe("tagger")
    transformer = nlp.get_pipe("transformer")
    tagger_trf = tagger.model.get_ref("tok2vec").layers[0]
    assert isinstance(transformer, Transformer)
    assert isinstance(tagger_trf, TransformerListener)
    assert tagger_trf.upstream_name == "custom_upstream"
    train_examples = []
    for t in TRAIN_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))
        for tag in t[1]["tags"]:
            tagger.add_label(tag)

    # Check that the Transformer component finds it listeners
    assert transformer.listeners == []
    optimizer = nlp.initialize(lambda: train_examples)
    assert tagger_trf in transformer.listeners

    for i in range(2):
        losses = {}
        nlp.update(train_examples, sgd=optimizer, losses=losses)

    text = "We're interested at underwater basket weaving."
    doc = nlp(text)
    doc_tensor = tagger_trf.predict([doc])
    _assert_equal_tensors(doc._.trf_data.tensors, doc_tensor[0].tensors)

    # ensure IO goes OK
    with make_tempdir() as d:
        file_path = d / "trained_nlp"
        nlp.to_disk(file_path)
        nlp2 = util.load_model_from_path(file_path)
        doc2 = nlp2(text)
        tagger2 = nlp2.get_pipe("tagger")
        tagger_trf2 = tagger2.model.get_ref("tok2vec").layers[0]
        doc_tensor2 = tagger_trf2.predict([doc2])
        _assert_equal_tensors(doc_tensor2[0].tensors, doc_tensor[0].tensors)

        # make sure that this can be saved to directory once more
        file_path_2 = d / "trained_nlp_2"
        nlp2.to_disk(file_path_2)

    # ensure to_bytes / from_bytes works
    nlp_bytes = nlp.to_bytes()
    nlp3 = util.load_model_from_config(orig_config, auto_fill=True, validate=True)
    nlp3.from_bytes(nlp_bytes)
    doc3 = nlp3(text)
    tagger3 = nlp3.get_pipe("tagger")
    tagger_trf3 = tagger3.model.get_ref("tok2vec").layers[0]
    doc_tensor3 = tagger_trf3.predict([doc3])
    _assert_equal_tensors(doc_tensor3[0].tensors, doc_tensor[0].tensors)


def test_transformer_sentencepiece_IO():
    """Test that a transformer using sentencepiece trains + IO goes OK"""
    orig_config = Config().from_str(cfg_string)
    orig_config["components"]["transformer"]["model"]["name"] = "camembert-base"
    nlp = util.load_model_from_config(orig_config, auto_fill=True, validate=True)
    tagger = nlp.get_pipe("tagger")
    tagger_trf = tagger.model.get_ref("tok2vec").layers[0]
    train_examples = []
    for t in TRAIN_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))
        for tag in t[1]["tags"]:
            tagger.add_label(tag)

    optimizer = nlp.initialize(lambda: train_examples)
    for i in range(2):
        losses = {}
        nlp.update(train_examples, sgd=optimizer, losses=losses)

    text = "We're interested at underwater basket weaving."
    doc = nlp(text)
    doc_tensor = tagger_trf.predict([doc])

    # ensure IO goes OK
    with make_tempdir() as d:
        file_path = d / "trained_nlp"
        nlp.to_disk(file_path)
        nlp2 = util.load_model_from_path(file_path)
        doc2 = nlp2(text)
        tagger2 = nlp2.get_pipe("tagger")
        tagger_trf2 = tagger2.model.get_ref("tok2vec").layers[0]
        doc_tensor2 = tagger_trf2.predict([doc2])
        _assert_equal_tensors(doc_tensor2[0].tensors, doc_tensor[0].tensors)

        # make sure that this can be saved to directory once more
        file_path_2 = d / "trained_nlp_2"
        nlp2.to_disk(file_path_2)

    # ensure to_bytes / from_bytes works
    nlp_bytes = nlp.to_bytes()
    nlp3 = util.load_model_from_config(orig_config, auto_fill=True, validate=True)
    nlp3.from_bytes(nlp_bytes)
    doc3 = nlp3(text)
    tagger3 = nlp3.get_pipe("tagger")
    tagger_trf3 = tagger3.model.get_ref("tok2vec").layers[0]
    doc_tensor3 = tagger_trf3.predict([doc3])
    _assert_equal_tensors(doc_tensor3[0].tensors, doc_tensor[0].tensors)


def test_transformer_pipeline_empty():
    """Test that the pipeline doesn't fail with empty input"""
    orig_config = Config().from_str(cfg_string)
    nlp = util.load_model_from_config(orig_config, auto_fill=True, validate=True)
    tagger = nlp.get_pipe("tagger")
    train_examples = []
    for t in TRAIN_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))
        for tag in t[1]["tags"]:
            tagger.add_label(tag)

    # train on empty doc
    optimizer = nlp.initialize()
    losses = {}
    empty_train_example = Example.from_dict(nlp.make_doc(""), {})
    nlp.update(train_examples, sgd=optimizer, losses=losses)
    nlp.update([empty_train_example], sgd=optimizer, losses=losses)
    train_examples.append(empty_train_example)
    nlp.update(train_examples, sgd=optimizer, losses=losses)

    # predict empty doc
    doc = nlp("")
    _assert_empty(doc._.trf_data)
    docs = nlp.pipe(["", ""])
    for doc in docs:
        _assert_empty(doc._.trf_data)
    nlp.pipe([])

    # predict combination of empty and non-empty
    doc = nlp("This is a sentence")
    normal_tags = [t.tag_ for t in doc]

    docs = list(nlp.pipe(["", "This is a sentence", "", ""]))
    _assert_empty(docs[0]._.trf_data)
    assert [t.tag_ for t in docs[0]] == []
    assert [t.tag_ for t in docs[1]] == normal_tags
    _assert_empty(docs[2]._.trf_data)
    _assert_empty(docs[3]._.trf_data)


def _assert_empty(trf_data):
    assert trf_data.wordpieces.strings == []
    assert trf_data.wordpieces.input_ids.size == 0
    assert trf_data.wordpieces.attention_mask.size == 0
    assert trf_data.tensors == ()
    assert len(trf_data.align.data) == 0


def test_replace_listeners():
    orig_config = Config().from_str(cfg_string)
    nlp = util.load_model_from_config(orig_config, auto_fill=True, validate=True)
    text = "This is awesome"
    examples = [Example.from_dict(nlp.make_doc(text), {"tags": ["A", "B", "C"]})]
    optimizer = nlp.initialize(lambda: examples)
    # verify correct configuration with transformer listener
    transformer = nlp.get_pipe("transformer")
    tagger = nlp.get_pipe("tagger")
    tagger_tok2vec = tagger.model.get_ref("tok2vec")
    tagger_listener = tagger_tok2vec.get_ref("listener")
    assert isinstance(tagger_listener, TransformerListener)
    assert transformer.listener_map["tagger"][0] == tagger_listener
    assert isinstance(transformer.model, TransformerModel)
    assert (
        nlp.config["components"]["transformer"]["model"]["@architectures"]
        == "spacy-transformers.TransformerModel.v3"
    )
    assert (
        nlp.config["components"]["tagger"]["model"]["tok2vec"]["@architectures"]
        == "spacy-transformers.TransformerListener.v1"
    )
    # train pipe before replacing listeners
    for i in range(2):
        losses = {}
        nlp.update(examples, sgd=optimizer, losses=losses)
        doc = nlp(text)

    preds = [t.tag_ for t in doc]
    doc_tensor = tagger_tok2vec.predict([doc])

    # replace listener and verify predictions are still the same
    nlp.replace_listeners("transformer", "tagger", ["model.tok2vec"])
    tagger = nlp.get_pipe("tagger")
    tagger_tok2vec = tagger.model.get_ref("tok2vec")
    assert isinstance(tagger_tok2vec, Model)
    assert tagger_tok2vec.layers[0].layers[0].name == "transformer"
    assert (
        nlp.config["components"]["tagger"]["model"]["tok2vec"]["@architectures"]
        == "spacy-transformers.Tok2VecTransformer.v3"
    )
    doc2 = nlp(text)
    assert preds == [t.tag_ for t in doc2]
    pred_tensor = tagger_tok2vec.predict([doc2])
    _assert_equal_tensors(doc_tensor, pred_tensor)

    # attempt training with the new pipeline
    optimizer = nlp.resume_training()
    for i in range(2):
        losses = {}
        nlp.update(examples, sgd=optimizer, losses=losses)
        assert losses["tagger"] > 0.0

    # check for presence of additional fields in model_output
    assert doc2._.trf_data.model_output.pooler_output is not None
    assert doc2._.trf_data.model_output.attentions is not None

    # ensure IO goes OK
    doc_tensor_trained = tagger_tok2vec.predict([doc])
    with make_tempdir() as d:
        file_path = d / "trained_nlp"
        nlp.to_disk(file_path)
        nlp2 = util.load_model_from_path(file_path)
        doc3 = nlp2(text)
        tagger2 = nlp2.get_pipe("tagger")
        tagger_tok2vec2 = tagger2.model.get_ref("tok2vec")
        pred_tensor = tagger_tok2vec2.predict([doc3])
        _assert_equal_tensors(doc_tensor_trained, pred_tensor)


def test_replace_listeners_invalid():
    orig_config = Config().from_str(cfg_string)
    nlp = util.load_model_from_config(orig_config, auto_fill=True, validate=True)
    text = "This is awesome"
    examples = [Example.from_dict(nlp.make_doc(text), {"tags": ["A", "B", "C"]})]
    optimizer = nlp.initialize(lambda: examples)
    for i in range(2):
        losses = {}
        nlp.update(examples, sgd=optimizer, losses=losses)
    with pytest.raises(ValueError):
        nlp.replace_listeners("invalid", "tagger", ["model.tok2vec"])
    with pytest.raises(ValueError):
        nlp.replace_listeners("transformer", "parser", ["model.tok2vec"])
    with pytest.raises(ValueError):
        nlp.replace_listeners("transformer", "tagger", ["model.yolo"])
    with pytest.raises(ValueError):
        nlp.replace_listeners("transformer", "tagger", ["model.tok2vec", "model.yolo"])


@pytest.fixture
def texts():
    data = [
        "Hello world.",
        "This is spacy.",
        "You can use multiprocessing with pipe method.",
        "Please try!",
    ]
    return data


def test_multiprocessing(simple_nlp, texts):
    ops = get_current_ops()
    if isinstance(ops, NumpyOps):
        texts = texts * 3
        expecteds = [simple_nlp(text) for text in texts]
        docs = simple_nlp.pipe(texts, n_process=2, batch_size=2)

        for doc, expected_doc in zip(docs, expecteds):
            assert_docs_equal(doc, expected_doc)



import pytest
from spacy.lang.en import English

from ..span_getters import configure_strided_spans, configure_get_sent_spans


@pytest.mark.parametrize(
    "window,stride,docs,result",
    [
        (4, 3, ["0", "1234", "56789a"], [["0"], ["1234"], ["5678", "89a"]]),
        (4, 4, ["0", "1234", "56789a"], [["0"], ["1234"], ["5678", "9a"]]),
        (4, 2, ["0", "1234", "56789a"], [["0"], ["1234"], ["5678", "789a"]]),
    ],
)
def test_get_strided_spans(window, stride, docs, result):
    get_strided = configure_strided_spans(window, stride)
    spans = get_strided(docs)
    assert spans == result


def test_get_sent_spans():
    nlp = English()
    nlp.add_pipe("sentencizer")
    doc = nlp("One. One more. Three sentences in total.")
    assert len(list(doc.sents)) == 3
    get_sent_spans = configure_get_sent_spans()
    spans = get_sent_spans([doc])[0]
    assert len(spans) == 3
    assert spans[0].text == "One."
    assert spans[1].text == "One more."
    assert spans[2].text == "Three sentences in total."


def test_get_custom_spans():
    def configure_custom_sent_spans(max_length: int):
        def get_custom_sent_spans(docs):
            spans = []
            for doc in docs:
                spans.append([])
                for sent in doc.sents:
                    start = 0
                    end = max_length
                    while end <= len(sent):
                        spans[-1].append(sent[start:end])
                        start += max_length
                        end += max_length
                    if start < len(sent):
                        spans[-1].append(sent[start : len(sent)])
            return spans

        return get_custom_sent_spans

    nlp = English()
    nlp.add_pipe("sentencizer")
    doc = nlp(
        "One. And one more. So that makes three sentences and this one is a bit longer."
    )
    assert len(list(doc.sents)) == 3
    get_sent_spans = configure_custom_sent_spans(max_length=4)
    spans = get_sent_spans([doc])[0]
    assert len(spans) == 6
    assert spans[0].text == "One."
    assert spans[1].text == "And one more."
    assert spans[2].text == "So that makes three"
    assert spans[3].text == "sentences and this one"
    assert spans[4].text == "is a bit longer"
    assert spans[5].text == "."



import pytest
from spacy_transformers.util import huggingface_from_pretrained
from spacy_transformers.util import huggingface_tokenize


def test_deprecation_warnings():
    with pytest.warns(DeprecationWarning):
        tokenizer, transformer = huggingface_from_pretrained(
            "distilbert-base-uncased", {}
        )
    with pytest.warns(DeprecationWarning):
        token_data = huggingface_tokenize(tokenizer, ["a", "b", "c"])



from typing import Any, Dict
from io import BytesIO
from pathlib import Path
import srsly
import torch
from thinc.api import get_current_ops
from spacy.util import SimpleFrozenDict

from ..data_classes import HFObjects
from ..util import make_tempdir

from thinc.api import PyTorchGradScaler, PyTorchShim

from transformers import AutoModel, AutoConfig, AutoTokenizer


class HFShim(PyTorchShim):
    """Interface between a HF Pytorch model and a Thinc Model."""

    def __init__(
        self,
        model: HFObjects,
        config=None,
        optimizer: Any = None,
        mixed_precision: bool = False,
        grad_scaler_config: dict = {},
    ):
        self._hfmodel = model

        # Enable gradient scaling when mixed precision is enabled and gradient
        # scaling is not explicitly disabled in the configuration.
        if "enabled" not in grad_scaler_config:
            grad_scaler_config["enabled"] = mixed_precision

        super().__init__(
            model.transformer,
            config,
            optimizer,
            mixed_precision,
            grad_scaler=PyTorchGradScaler(**grad_scaler_config),
        )

    def to_bytes(self):
        config = {}
        tok_dict = {}
        weights_bytes = {}
        tok_cfg = {}
        trf_cfg = {}
        hf_model = self._hfmodel
        if hf_model.transformer is not None:
            tok_dict = {}
            config = hf_model.transformer.config.to_dict()
            tokenizer = hf_model.tokenizer
            with make_tempdir() as temp_dir:
                if hasattr(tokenizer, "vocab_file"):
                    vocab_file_name = tokenizer.vocab_files_names["vocab_file"]
                    vocab_file_path = str((temp_dir / vocab_file_name).absolute())
                    with open(vocab_file_path, "wb") as fileh:
                        fileh.write(hf_model.vocab_file_contents)
                    tokenizer.vocab_file = vocab_file_path
                tokenizer.save_pretrained(str(temp_dir.absolute()))
                for x in temp_dir.glob("**/*"):
                    if x.is_file():
                        tok_dict[x.name] = x.read_bytes()
            filelike = BytesIO()
            torch.save(self._model.state_dict(), filelike)
            filelike.seek(0)
            weights_bytes = filelike.getvalue()
        else:
            tok_cfg = hf_model._init_tokenizer_config
            trf_cfg = hf_model._init_transformer_config
        msg = {
            "config": config,
            "state": weights_bytes,
            "tokenizer": tok_dict,
            "_init_tokenizer_config": tok_cfg,
            "_init_transformer_config": trf_cfg,
        }
        return srsly.msgpack_dumps(msg)

    def from_bytes(self, bytes_data):
        msg = srsly.msgpack_loads(bytes_data)
        config_dict = msg["config"]
        tok_dict = msg["tokenizer"]
        if config_dict:
            with make_tempdir() as temp_dir:
                config_file = temp_dir / "config.json"
                srsly.write_json(config_file, config_dict)
                config = AutoConfig.from_pretrained(config_file)
                for x, x_bytes in tok_dict.items():
                    Path(temp_dir / x).write_bytes(x_bytes)
                tokenizer = AutoTokenizer.from_pretrained(str(temp_dir.absolute()))
                vocab_file_contents = None
                if hasattr(tokenizer, "vocab_file"):
                    vocab_file_name = tokenizer.vocab_files_names["vocab_file"]
                    vocab_file_path = str((temp_dir / vocab_file_name).absolute())
                    with open(vocab_file_path, "rb") as fileh:
                        vocab_file_contents = fileh.read()

            transformer = AutoModel.from_config(config)
            self._hfmodel = HFObjects(
                tokenizer,
                transformer,
                vocab_file_contents,
                SimpleFrozenDict(),
                SimpleFrozenDict(),
            )
            self._model = transformer
            filelike = BytesIO(msg["state"])
            filelike.seek(0)
            ops = get_current_ops()
            if ops.device_type == "cpu":
                map_location = "cpu"
            else:  # pragma: no cover
                device_id = torch.cuda.current_device()
                map_location = f"cuda:{device_id}"
            self._model.load_state_dict(torch.load(filelike, map_location=map_location))
            self._model.to(map_location)
        else:
            self._hfmodel = HFObjects(
                None,
                None,
                None,
                msg["_init_tokenizer_config"],
                msg["_init_transformer_config"],
            )
        return self
